// EMOD test


#include "hp3000_defs.h"
#include "sim_defs.h"


#pragma GCC diagnostic ignored "-Wunused-function"



static uint32 dev;
static uint32 unit [2];

static struct {
    UNIT   *uptr;
    int32  value;
    char   *cptr;
    char   char1;
    void   *desc;
    t_stat tstat;
    } param;


static t_stat emod_reset (DEVICE *dptr);
static t_stat setup      (UNIT *uptr, int32 value, char *cptr, void *desc, t_stat status);

static t_stat set_ok   (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat set_fail (UNIT *uptr, int32 value, char *cptr, void *desc);
static t_stat set_skip (UNIT *uptr, int32 value, char *cptr, void *desc);

static t_stat show_ok   (FILE *st, UNIT *uptr, int32 value, void *desc);
static t_stat show_fail (FILE *st, UNIT *uptr, int32 value, void *desc);
static t_stat show_skip (FILE *st, UNIT *uptr, int32 value, void *desc);


extern DEVICE emod_dev;


static UNIT emod_unit [] = {
    { UDATA (NULL, UNIT_DISABLE, 0), 0 },
    { UDATA (NULL, UNIT_DISABLE, 0), 0 },
    };


static REG emod_reg [] = {
/*    Macro   Name      Location              Radix  Width  Offset  Depth   Flags          */
/*    ------  --------  --------------------  -----  -----  ------  -----  --------------- */
    { GRDATA (DFLAGS,   emod_dev.flags,        16,    32,      0),         PV_RZRO         },
    { GRDATA (DCTRL,    emod_dev.dctrl,        16,    32,      0),         PV_RZRO         },
    { URDATA (UFLAGS,   emod_unit [0].flags,   16,    32,      0,     2,   PV_RZRO)        },
    { URDATA (U4,       emod_unit [0].u4,      16,    32,      0,     2,   PV_RZRO)        },

    { GRDATA (DGLOBAL,  dev,                   16,    32,      0),         PV_RZRO         },
    { BRDATA (UGLOBAL,  unit,                  16,    32,             2),  PV_RZRO         },

    { GRDATA (UPTR,     param.uptr,            16,    32,      0),         PV_RZRO         },
    { GRDATA (VALUE,    param.value,            8,    32,      0),         PV_LEFT         },
    { GRDATA (CPTR,     param.cptr,            16,    32,      0),         PV_RZRO         },
    { GRDATA (CHAR1,    param.char1,            8,     8,      0),         PV_LEFT         },
    { GRDATA (DESC,     param.desc,            16,    32,      0),         PV_RZRO         },
    { GRDATA (TSTAT,    param.tstat,            8,    16,      0),         PV_RZRO         },

    { NULL }
    };


#define D1          MTAB_EDV                                    /* device flags field */
#define D2          MTAB_EDV | MTAB_EGL                         /* device scalar */
#define D3          MTAB_EDV | MTAB_EMK                         /* device flags field, mask value */

#define U1          MTAB_EUN                                    /* unit field */
#define U2          MTAB_EUN | MTAB_EGL                         /* unit array */

#define F1          UNIT_UFMASK | UNIT_DISABLE

#define V1          D2 | MTAB_ERX (2)                               /* device scalar, radix 2 */
#define V2          D2 | MTAB_ERX (16)                              /* device scalar, radix 16 */
#define V3          D2 | MTAB_ERX (10) | MTAB_EBT                   /* device scalar, radix 10, RL position */
#define V4          D2 | MTAB_ERX (10) | MTAB_EBT | MTAB_ELR        /* device scalar, radix 10, LR position */
#define V5          D2 | MTAB_ERX (10) | MTAB_EBI | MTAB_EMN (9)    /* device scalar, radix 10, bias by 9 */
#define V6          V3 | MTAB_EBI | MTAB_EMN (9)                    /* device scalar, radix 10, RL position, bias by 9 */
#define V7          V4 | MTAB_EBI | MTAB_EMN (9)                    /* device scalar, radix 10, LR position, bias by 9 */
#define V8          D2 | MTAB_ERX (10)                              /* device scalar, radix 10 */

#define E1          MTAB_EDV | MTAB_EUN                         /* error: both D and U bits set */
#define E2          MTAB_EDV | MTAB_EGL                         /* error: G bit set with NULL descriptor */
#define E3          E2 | MTAB_ERX (2) | MTAB_ERO                /* device scalar, radix 2, read-only */
#define E4          E2 | MTAB_ERX (2)                           /* device scalar, radix 2 */
#define E5          E2 | MTAB_ERX (10) | MTAB_EMN (5)           /* device scalar value, radix 10, minimum 5 */

#define S0          MTAB_EDV | MTAB_EGL | MTAB_ENO                  /* device scalar, named-only */
#define S1          S0 | MTAB_ERX (10)                              /* device scalar, radix 10 */
#define S2          S0 | MTAB_ERX (2)                               /* device scalar, radix 2 */
#define S3          S0 | MTAB_ERX (10) | MTAB_EBT                   /* device scalar, radix 10, RL position */
#define S4          S0 | MTAB_ERX (10) | MTAB_EBT | MTAB_ELR        /* device scalar, radix 10, LR position */
#define S5          S0 | MTAB_ERX (10) | MTAB_EBI | MTAB_EMN (9)    /* device scalar, radix 10, bias by 9 */
#define S6          S3 | MTAB_EBI | MTAB_EMN (9)                    /* device scalar, radix 10, RL position, bias by 9 */
#define S7          S4 | MTAB_EBI | MTAB_EMN (9)                    /* device scalar, radix 10, LR position, bias by 9 */

#define BF_0400     BITS ( 4,  0)
#define BF_1508     BITS (15,  8)
#define BF_1904     BITS (19,  4)
#define BF_2724     BITS (27, 24)
#define BF_3124     BITS (31, 24)

#define FLAG_00     (1u <<  0)
#define FLAG_15     (1u << 15)
#define FLAG_16     (1u << 16)
#define FLAG_30     (1u << 30)
#define FLAG_31     (1u << 31)


static MTAB emod_mod [] = {
/*    Mask       Match/Max  Print    Match                                                     Expansion */
/*    Value      Value      String   String   Validation   Display     Descriptor          4.x    Flags     */
/*    --------  ----------  -------  -------  -----------  ----------  ------------------  -----  --------- */

// Section 1

    { ~0u,     ~0u,         "DFX",   "DFX",   NULL,        NULL,       NULL,               NULL,  D1        },
    { ~0u,     ~0u,         "DUX",   "DUX",   NULL,        NULL,       &emod_dev.dctrl,    NULL,  D1        },
    { ~0u,     ~0u,         "DGX",   "DGX",   NULL,        NULL,       &dev,               NULL,  D2        },

    { ~0u,     ~0u,         "UFX",   "UFX",   NULL,        NULL,       NULL,               NULL,  U1        },
    { ~0u,     ~0u,         "UUX",   "UUX",   NULL,        NULL,       &emod_unit [0].u4,  NULL,  U1        },
    { ~0u,     ~0u,         "UGX",   "UGX",   NULL,        NULL,       &unit,              NULL,  U2        },

    { ~0u,      F1,         "UFX",   NULL,    NULL,        NULL,       NULL,               NULL,  U1        },

// Section 2

    { FLAG_00,  FLAG_00,    "DS00",  "DS00",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_15,  FLAG_15,    "DS15",  "DS15",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_16,  FLAG_16,    "DS16",  "DS16",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_30,  FLAG_30,    "DS30",  "DS30",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_31,  FLAG_31,    "DS31",  "DS31",  NULL,        NULL,       NULL,               NULL,  D1        },

    { FLAG_00,  0,          "DC00",  "DC00",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_15,  0,          "DC15",  "DC15",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_16,  0,          NULL,    "DC16",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_30,  0,          NULL,    "DC30",  NULL,        NULL,       NULL,               NULL,  D1        },
    { FLAG_31,  0,          "DC31",  "DC31",  NULL,        NULL,       NULL,               NULL,  D1        },

    { FLAG_30,  FLAG_30,    NULL,    "DV30",  set_ok,      NULL,       NULL,               NULL,  D1        },
    { FLAG_30,  0,          NULL,    "DM30",  set_skip,    NULL,       NULL,               NULL,  D1        },
    { FLAG_30,  0,          NULL,    "DK30",  set_ok,      NULL,       NULL,               NULL,  D3        },

// Section 3

    { BF_1508,  255,        NULL,    "V1",    NULL,        NULL,       &dev,               NULL,  V1        },
    { BF_1508,  255,        NULL,    "V2",    set_ok,      NULL,       &dev,               NULL,  V2        },
    { BF_1904,   15,        NULL,    "V3",    set_ok,      NULL,       &dev,               NULL,  V3        },
    { BF_1904,   15,        NULL,    "V4",    set_ok,      NULL,       &dev,               NULL,  V4        },
    { BF_2724,   15,        NULL,    "V5",    set_ok,      NULL,       &dev,               NULL,  V5        },
    { BF_3124,   16,        NULL,    "V6",    set_ok,      NULL,       &dev,               NULL,  V6        },
    { BF_3124,   16,        NULL,    "V7",    set_ok,      NULL,       &dev,               NULL,  V7        },
    { BF_0400,   31,        NULL,    "V8",    set_skip,    NULL,       &dev,               NULL,  V8        },

// Section 4

    { FLAG_16,  FLAG_16,    NULL,    "E1",    NULL,        NULL,       NULL,               NULL,  E1        },
    { FLAG_16,  FLAG_16,    NULL,    "E2",    NULL,        NULL,       NULL,               NULL,  E2        },
    { BF_1508,    5,        NULL,    "E3",    NULL,        NULL,       &dev,               NULL,  E5        },
    { BF_1508,  255,        NULL,    "E4",    NULL,        NULL,       &dev,               NULL,  E3        },
    { BF_1508,  255,        NULL,    "E5",    NULL,        NULL,       &dev,               NULL,  E4        },
    { BF_1508,  255,        NULL,    "E6",    NULL,        NULL,       &dev,               NULL,  E5        },
    { BF_1508,  255,        NULL,    "E7",    set_fail,    NULL,       &dev,               NULL,  E5        },
    { BF_1508,  255,        "E8=",   "E8",    NULL,        show_fail,  &dev,               NULL,  E3        },

// Section 6

    { BF_2724,   15,        "S1=",   "S1",    NULL,        NULL,       &dev,               NULL,  S1        },
    { BF_2724,   15,        "S2=",   "S2",    NULL,        NULL,       &dev,               NULL,  S2        },
    { BF_1904,   15,        "S3=",   "S3",    NULL,        show_ok,    &dev,               NULL,  S3        },
    { BF_1904,   15,        "S4=",   "S4",    NULL,        show_ok,    &dev,               NULL,  S4        },
    { BF_2724,   15,        "S5=",   "S5",    NULL,        show_ok,    &dev,               NULL,  S5        },
    { BF_1904,   15,        "S6=",   "S6",    NULL,        show_ok,    &dev,               NULL,  S6        },
    { BF_3124,   16,        "S7=",   "S7",    NULL,        show_ok,    &dev,               NULL,  S7        },
    { BF_2724,   15,        "S8=",   "S8",    NULL,        show_skip,  &dev,               NULL,  S1        },
    { BF_1904,   15,        "S9=",   "S9",    NULL,        show_ok,    &dev,               NULL,  S3        },

    { 0 }
    };


DEVICE emod_dev = {
    "EMOD",                                     /* device name */
    emod_unit,                                  /* unit array */
    emod_reg,                                   /* register array */
    emod_mod,                                   /* modifier array */
    2,                                          /* number of units */
    10,                                         /* address radix */
    32,                                         /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    8,                                          /* data width */
    NULL,                                       /* examine routine */
    NULL,                                       /* deposit routine */
    emod_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    NULL,                                       /* attach routine */
    NULL,                                       /* detach routine */
    NULL,                                       /* device information block pointer */
    0,                                          /* device flags */
    0,                                          /* debug control flags */
    NULL,                                       /* debug flag name array */
    NULL,                                       /* memory size change routine */
    NULL,                                       /* logical device name */
    NULL,                                       /* (4.0 dummy) help routine */
    NULL,                                       /* (4.0 dummy) help attach routine */
    NULL,                                       /* (4.0 dummy) help context */
    NULL,                                       /* (4.0 dummy) description */
    NULL                                        /* the device adjustment function pointer */
    };


static t_stat emod_reset (DEVICE *dptr)
{
dev = 0;
unit [0] = unit [1] = 0;

emod_dev.flags      &= 0x0000FFFF;
emod_unit [0].flags &= 0x0000FFFF;
emod_unit [1].flags &= 0x0000FFFF;

emod_dev.dctrl   = 0;
emod_unit [0].u4 = 0;
emod_unit [1].u4 = 0;

return SCPE_OK;
}


static t_stat setup (UNIT *uptr, int32 value, char *cptr, void *desc, t_stat status)
{
param.uptr  = uptr; 
param.value = value;
param.cptr  = cptr; 
param.desc  = desc; 

if (cptr != NULL)
    param.char1 = *cptr;
else
    param.char1 = '\0';

param.tstat = status;

return status;
}

static t_stat set_ok (UNIT *uptr, int32 value, char *cptr, void *desc)
{
return setup (uptr, value, cptr, desc, SCPE_OK);
}


static t_stat set_fail (UNIT *uptr, int32 value, char *cptr, void *desc)
{
return setup (uptr, value, cptr, desc, SCPE_ARG);
}


static t_stat set_skip (UNIT *uptr, int32 value, char *cptr, void *desc)
{
return setup (uptr, value, cptr, desc, -1);
}


static t_stat show_ok   (FILE *st, UNIT *uptr, int32 value, void *desc)
{
return setup (uptr, value, NULL, desc, SCPE_OK);
}


static t_stat show_fail (FILE *st, UNIT *uptr, int32 value, void *desc)
{
return setup (uptr, value, NULL, desc, SCPE_NOFNC);
}


static t_stat show_skip (FILE *st, UNIT *uptr, int32 value, void *desc)
{
return setup (uptr, value, NULL, desc, -1);
}
